// TweeningDemo.cpp
// by Brian Jorgensen
// bjmumblingmiles@hotmail.com
// You must have the DX9 SDK to compile this project.

// DISCLAIMER
// I am not responsible for anything...ever... ;)

#include <fstream>
#include <windows.h>
#include "Window.h"
#include <d3d9.h>
#include <d3dx9.h>
#include "SafeRelease.h"
#include "MD2Model.h"

#include "TweeningDemo.h"

// CLEAR COLOR - press c to cycle through this list
const D3DCOLOR clear_color[] = 
{
	D3DCOLOR_XRGB(0,0,0), // 0
	D3DCOLOR_XRGB(0,0,255),
	D3DCOLOR_XRGB(0,127,255), // 2
	D3DCOLOR_XRGB(0,255,0),
	D3DCOLOR_XRGB(255,255,0), // 4
	D3DCOLOR_XRGB(255,0,0),
	D3DCOLOR_XRGB(255,0,255), // 6
	D3DCOLOR_XRGB(255,255,255),
	D3DCOLOR_XRGB(127,127,127) // 8
};
const int CLEAR_COLOR_AMT = 9;
int current_color = 0;

// variables
IDirect3D9* d3d = 0;
IDirect3DDevice9* d3d_device = 0;

IDirect3DVertexDeclaration9* tween_decl = 0;
IDirect3DTexture9* md2_texture = 0;
CMD2Model *md2_model = 0;

float rotation_factor = 0.0f;
float zoom_factor = 0.0f;

// functions
void InitScene(void)
{
	// For this example, I'm using an MD2 loader I wrote
	// If you just want to understand tweening,
	// just look at the model class as a source of vertex
	// buffers
	md2_model = new CMD2Model;

	md2_model->LoadFromFile("Data\\tris.md2", d3d_device);
	if (!md2_model->IsValid())
	{
		MessageBox(0,"Loading model failed. Looking for data at \\data","Fatal Error",0);
		PostQuitMessage(0);
	}

	// Set up texture stuff

	D3DXCreateTextureFromFile(d3d_device, "Data\\hobgoblin.bmp", &md2_texture);
	if (!md2_texture)
	{
		MessageBox(0,"Loading texture failed. Looking for data at \\data","Fatal Error",0);
		PostQuitMessage(0);
	}

	d3d_device->SetTextureStageState(0,D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
	d3d_device->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
	d3d_device->SetTextureStageState(0,D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

	/////////////////////////////
	// Set up the declaration: //
	/////////////////////////////

	D3DVERTEXELEMENT9 tween_decl_ve[] =
	{
		{0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
		{1, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 1},
		{2, 0, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
		D3DDECL_END() // this macro is always needed as the last item! DON'T FORGET ;)
	};
	d3d_device->CreateVertexDeclaration(tween_decl_ve, &tween_decl);
	//                         //
	/////////////////////////////
	if (!tween_decl)
	{
		MessageBox(0,"Creating vertex declaration failed.","Fatal Error",0);
		PostQuitMessage(0);
	}
}

void RenderScene(void)
{
	if (GetAsyncKeyState(VK_LEFT))
		rotation_factor-=0.02f;
	if (GetAsyncKeyState(VK_RIGHT))
		rotation_factor+=0.02f;
	if (GetAsyncKeyState(VK_DOWN))
		zoom_factor-=1.0f;
	if (GetAsyncKeyState(VK_UP))
		zoom_factor+=1.0f;
	
	// Ahh...RenderScene. Here's where the action is...
	d3d_device->BeginScene();
	d3d_device->Clear(0, 0, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, clear_color[current_color], 1.0f, 0);
	
	// Set texture
	d3d_device->SetTexture(0, md2_texture);

	// Set transformations
	D3DXMATRIX mat_rot, mat_t, mat_user;

	// Fix orientation of model- md2's are sideways and upsidedown
	D3DXMatrixRotationYawPitchRoll(&mat_rot, 1.5f*D3DX_PI,1.5f*D3DX_PI, 0);
	// Rotate with arrow keys
	D3DXMatrixRotationY(&mat_user, rotation_factor*D3DX_PI);
	D3DXMatrixMultiply(&mat_rot, &mat_rot, &mat_user);
	d3d_device->SetTransform(D3DTS_WORLD, &(D3DMATRIX)mat_rot);
	// Backup viewer - zoom with arrow keys
	D3DXMatrixTranslation(&mat_t, 0.0f, 0.0f, -100.0f+zoom_factor);
	d3d_device->SetTransform(D3DTS_VIEW, &(D3DMATRIX)mat_t);

	///////////////////////////////////////////////////////////////
	// Here's the rendering code                                 //
	// It's set up to loop through all the frames of an md2 file //
	///////////////////////////////////////////////////////////////
	static float tween_factor = 0.0f;
	static int frame_1 = 0;
	static int frame_2 = 1;

	// Set this factor to increase/decrease the animation speed
	// If you are curious, try to make this time based- it's much better
	// I'm just leaving it incremental so it is easier to understand :)
	tween_factor += 0.1f;
	if (tween_factor >= 1.0f)
	{
		frame_1++;	// goto the next frame
		frame_2++;	//
		tween_factor = 0.0f; // we're tweening between two new frames now,
							 // so the tween_factor is reset

		// don't try to draw frames with non-existant buffers!
		if (frame_1 >= md2_model->GetFrameCount()) frame_1 = 0;
		if (frame_2 >= md2_model->GetFrameCount()) frame_2 = 0;
	}

	// turn on tweening, and set the tween factor
	d3d_device->SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_TWEENING);
	d3d_device->SetRenderState(D3DRS_TWEENFACTOR, FtoDW(tween_factor));

	// set our vertex declaration
	d3d_device->SetVertexDeclaration(tween_decl);

	// set our VB's
	d3d_device->SetIndices(md2_model->GetModelIB());
	d3d_device->SetStreamSource(0,md2_model->GetModelVB(frame_1),0,sizeof(MD2_POSITION_VERTEX));
	d3d_device->SetStreamSource(1,md2_model->GetModelVB(frame_2),0,sizeof(MD2_POSITION_VERTEX));
	d3d_device->SetStreamSource(2,md2_model->GetTextureVB(),0,sizeof(MD2_TEXTURE_VERTEX));

	// draw the tweened model!
	d3d_device->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, md2_model->GetVertCount(), 0, md2_model->GetTriangleCount());	

	// tidy up in case other types of objects are rendered (like a level)
	d3d_device->SetVertexDeclaration(0);
	d3d_device->SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_DISABLE);

	// present!
	d3d_device->EndScene();
	d3d_device->Present(0,0,0,0);

	d3d_device->SetTexture(0, 0);

	//////////////////////////////////////////////////////////////////////
	// NOTE                                                            //
	//////////////////////////////////////////////////////////////////////
	// Here's an exercise for the reader: make this time-based          //
	// I have it time-based in my in-progress engine, and it is         //
	// beautiful! MD2 animations are made at different fps'             //
	// Another fun idea would be to make lots of guys standing in lines //
	// it looks like an army- or an aerobics class...depending on the   //
	// md2                                                              //
	//////////////////////////////////////////////////////////////////////
}

void CleanupScene(void)
{
	delete md2_model;
	
	SafeRelease(tween_decl);
	SafeRelease(md2_texture);
}

void MatrixProjection(D3DMATRIX& mat, int width, int height, float fov, float near_plane, float far_plane)
{
	const float fAspect = (float)(((float)height)/(float)width);

	const float w = fAspect * (float)(cosf(fov/2.0f)/sinf(fov/2.0f));
	const float h = 1.0f * (float)(cosf(fov/2.0f)/sinf(fov/2.0f));
	const float Q = far_plane / (far_plane - near_plane);

	ZeroMemory(&mat,sizeof(mat));
	mat.m[0][0] = w;
	mat.m[1][1] = h;
	mat.m[2][2] = Q;
	mat.m[2][3] = 1.0f;
	mat.m[3][2] = -Q*near_plane;

	mat._33 = -mat._33;
	mat._34 = -mat._34;
}

D3DFORMAT FindPixelFormat(IDirect3D9* d3d)
{
	D3DFORMAT format;

	format = D3DFMT_A8R8G8B8;
	if (SUCCEEDED(d3d->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, format, format, false)))
		return format;
	
	format = D3DFMT_X8R8G8B8;
	if (SUCCEEDED(d3d->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, format, format, false)))
		return format;

	format = D3DFMT_R5G6B5;
	if (SUCCEEDED(d3d->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, format, format, false)))
		return format;
	
	format = D3DFMT_X1R5G5B5;
	if (SUCCEEDED(d3d->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, format, format, false)))
		return format;

	MessageBox(0,"FindFormat failed: No compatible backbuffer found!","Fatal Error",0);
	PostQuitMessage(0);
	return (D3DFORMAT)0;
}

void InitDirect3D(CWindow& window)
{
	d3d = Direct3DCreate9(D3D_SDK_VERSION);
	if (d3d == 0)
	{
		MessageBox(0,"Could not init Direct3D", "Fatal Error", 0);
		PostQuitMessage(0);
		return;
	}

	// Get display mode
	D3DDISPLAYMODE display_mode;
	d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,&display_mode);

	D3DPRESENT_PARAMETERS dpp;
	memset(&dpp,0,sizeof(D3DPRESENT_PARAMETERS));
	
	dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	dpp.Windowed = window.is_windowed();
	
	dpp.BackBufferCount = 1;

	if (dpp.Windowed)
	{
		dpp.BackBufferFormat = display_mode.Format;
		dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	}
	else
	{
		dpp.BackBufferWidth = window.getWidth();
		dpp.BackBufferHeight = window.getHeight();
		dpp.BackBufferFormat = FindPixelFormat(d3d);
		dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	}
	
	dpp.EnableAutoDepthStencil = true;
	dpp.AutoDepthStencilFormat = D3DFMT_D16;
	dpp.hDeviceWindow = 0;
	dpp.FullScreen_RefreshRateInHz = 0;
	dpp.Flags = 0;
	
	// Create device
	bool device_created = D3D_OK == d3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, (HWND)window, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &dpp, &d3d_device);
	if (d3d_device == 0 || device_created == 0)
	{
		MessageBox(0,"Could not init Direct3DDevice", "Fatal Error", 0);
		PostQuitMessage(0);
		return;
	}

	d3d_device->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	d3d_device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	d3d_device->SetRenderState(D3DRS_LIGHTING, FALSE);

	D3DMATRIX mat;
	MatrixProjection(mat, window.getWidth(), window.getHeight(), D3DX_PI/3, 1.0f, 2500.0f);
	d3d_device->SetTransform(D3DTS_PROJECTION, &mat);
}

void CleanupDirect3D(void)
{
	SafeRelease(d3d_device);
	SafeRelease(d3d);
}

int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE unused,
				   LPSTR cmd_line, int show_cmd)
{
	// So my window class collides with MFC?			
	// TIP: don't use them together :-p
	CWindow window(hinstance, "Brian's Tweening Demo", 800, 600, WS_OVERLAPPEDWINDOW, WindowProc);
	
	// If you want full screen, enable the code below and comment out the code above. Set your
	// resolution below
	//CWindow window(hinstance, "Brian's Tweening Demo", 800, 600, WS_POPUP, WindowProc);
	
	// Now, I initialize D3D in a non-OOP manner,
	// frankly because I don't want to give away
	// my messy Direct3D & Device classes.
	InitDirect3D(window);
	InitScene();

	MSG msg;
	while (1)
	{
		if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
				break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			RenderScene();
		}

		if (GetAsyncKeyState(VK_ESCAPE))
		{
			PostQuitMessage(0);
		}
	}
	CleanupScene();
	CleanupDirect3D();
	return 0;
}

LRESULT CALLBACK WindowProc(HWND hwnd, unsigned int msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	case WM_KEYDOWN:
		if (wparam == 'C')
		{
			current_color++;
			if (current_color >= CLEAR_COLOR_AMT)
				current_color = 0;
		}
		break;
	}
	return DefWindowProc(hwnd,msg,wparam,lparam);
}